//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	spec.cc -- implementation of the Spec class
//
//	CHANGE HISTORY:
//              DAH, 25/03/98: 1.0.0: - first version
//              DAH, 06/04/98: 1.0.1: - make _copyGlobalAttribs method
//
//	AUTHOR:
//              D. Hinshaw
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Expanded description of SAS task `evselect'
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/task_devel/packages/evselect/src/histo.cc,v 1.3 2000/12/22 13:36:12 ulammers Exp $

#include <iostream>
#include <cmath>
#include "param.h"
#include "Dal.h"
#include "errstr.h"
#include <ExprColumnGenerator.h>
#include "Extractor.h"
#include "DataSubspace.h"

void Histo::make(bool writeDss){

  bool zcol = booleanParameter("withzcolumn");

  // Get pointers to the needed columns

  Column *hDatCol, *zDatCol = 0, *zDatErrCol = 0;
  Column *hCol, *zCol, *zErrCol;

  string hcolumn = stringParameter("histogramcolumn");
  if( _intable->hasColumn(hcolumn) ){
    hCol = _intable->column(hcolumn);
    hDatCol = reType(hCol, Dal::Real64);
  }else{
    strstream newName;
    newName << "H" << _intable << ends;
    ExprColumnGenerator colGen(hcolumn, newName.str());
    colGen.doIt(*_intable, *_tempTable);
    hCol = hDatCol = _intable->column(hcolumn);
  }

  if(zcol){
    string zcolumn = stringParameter("zcolumn");
    if( _intable->hasColumn(zcolumn) ){
      zCol = _intable->column(zcolumn);
      zDatCol=reType(zCol, Dal::Real64);
    }else{
      strstream newName;
      newName << "Z" << _intable << ends;
      ExprColumnGenerator colGenZ(zcolumn, newName.str());
      colGenZ.doIt(*_intable, *_tempTable);
      zCol = zDatCol = _intable->column(zcolumn);
    }

    zcolumn = stringParameter("zerrorcolumn");
    if( _intable->hasColumn(zcolumn) ){
      zErrCol = _intable->column(zcolumn);
      zDatErrCol=reType(zErrCol, Dal::Real64);
    }else{
      strstream newName;
      newName << "ZE" << _intable << ends;
      ExprColumnGenerator colGenZ(zcolumn, newName.str());
      colGenZ.doIt(*_intable, *_tempTable);
      zErrCol = zDatErrCol = _intable->column(zcolumn);
    }
  }

  bool hIntCol = hCol->dataType()==Dal::Int8 || hCol->dataType()==Dal::Int16 || 
                 hCol->dataType()==Dal::Int32;

  // Read in the needed parameters

  _binSize = realParameter("histogrambinsize");
  if( _binSize==0 )
    errstr << "Histogram bin size set to zero.  This does not make sense." <<Sas::error("HistoZeroBinVal");


  if( hIntCol && floor(_binSize) != _binSize ){
    errstr << "Binning factor must be an integer for integer-valued columns."<<Sas::error("HistoBinVal");
  }

  // Determine the histogram range

  vector<real64> useRange(2);
  GetLimits(hCol, "histogrammin", "histogrammax", booleanParameter("withhistoranges"), useRange);

  if( hIntCol ){
    if( useRange[0] != floor(useRange[0]) || useRange[1] != floor(useRange[1]) ){
      errstr << "Ranges must be integers for integer-valued columns."<<Sas::error("HistoRangeVal");
    }
    useRange[0] -= 0.5;
    useRange[1] += 0.5;
  }

  if( useRange[1] == useRange[0] )
    errstr << "The column for extracting the histogram has "
	   << "the same value throughout, or the upper and lower limits "
	   << "for this column are equal. An empty histogram will "
	   << "be created." << Sas::warning("BadHistoLimits"); 
  if( useRange[1] < useRange[0] )
    errstr << "Upper histogram limit < lower limit.  Cannot continue." 
	   << Sas::error("BadHistoLimits");

  unsigned long nRows = (unsigned long)((useRange[1]-useRange[0])/_binSize);

  // Now initialize the output histogram

  Table * outTable = _dataSet->addTable("HISTO", nRows);

  _copyGlobalAttribs(outTable);

  // Make the necessary columns.
  
  real64 *cntDat, *cntErrDat= 0;
  real64 *zData = 0, *zErrData = 0;
  ColumnData *cErrCntDat = 0, *cZDatCol, *cZDatErrCol, *cCntDat;

  if( zcol ){
    cErrCntDat = outTable->addColumn("ERROR", Dal::Real64, "", "count")->data();
    cntErrDat = cErrCntDat->real64Data();
    for(unsigned i=0; i<nRows; i++) cntErrDat[i] = 0;

    cZDatCol = zDatCol->data();
    zData = cZDatCol->real64Data();
    cZDatErrCol = zDatErrCol->data();
    zErrData = cZDatErrCol->real64Data();

  }

  cCntDat = outTable->addColumn("COUNTS", Dal::Real64, "", "count")->data();
  cntDat = cCntDat->real64Data();
  for(unsigned i=0; i<nRows; i++) cntDat[i] = 0;

  real64 histoZero = useRange[0] + _binSize/2;

  ColumnData *cValueDat= outTable->addColumn(hcolumn, Dal::Real64, "", hCol->units())->data();
  real64 *valueDat = cValueDat->real64Data();
  for(unsigned i=0; i<nRows; i++) valueDat[i] = histoZero + i*_binSize;
  delete cValueDat;


  // Set up type converted columns

  Column *flagDatCol;

  ColumnData *cHDatCol = hDatCol->data(), *cFlagData = 0;
  int32 index, *flagData = 0;
  real64 *hData = cHDatCol->real64Data();

  if(! _destruct){
    flagDatCol = reType(_intable->column(_flagCol), Dal::Int32);
    cFlagData = flagDatCol->data();
    flagData = cFlagData->int32Data();
  }

  // Accumulate the columns

  unsigned long mRows = _intable->rows();
  for(unsigned long irow=0; irow<mRows; irow++){
    if( ! _destruct && ((flagData[irow] & _flagBit) == 0) ) continue;
    index = (int32)floor( float(hData[irow]-useRange[0])/_binSize );
    if( index >= 0 && index < (int32)nRows ){
      if( zcol ){
	cntDat[index] += zData[irow];
	cntErrDat[index] += zErrData[irow]*zErrData[irow];
      }else
	cntDat[index] += 1;
    }
  }

  // Final tidying up

  if( zcol ){
    for(unsigned long irow=0; irow<outTable->rows(); irow++)
      cntErrDat[irow] = sqrt(cntErrDat[irow]);

    delete cErrCntDat;
  }

  delete cCntDat;

  delete cHDatCol;

  if(! _destruct) delete cFlagData;

  // Write data subspace info
  if (writeDss) {
     DataSubspace dss(_intable);
     outTable << dss;
  }

  // Done
}

