//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	spec.cc -- implementation of the Spec class
//
//	CHANGE HISTORY:
//              DAH, 25/03/98: 1.0.0: - first version
//              DAH, 06/04/98: 1.0.1: - make _copyGlobalAttribs method
//				UL, 2000-12-20: 2.0.0: - couple of trivial changes to make the
//										 optimizer happy
//
//	AUTHOR:
//              D. Hinshaw
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Expanded description of SAS task `evselect'
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/task_devel/packages/evselect/src/spec.cc,v 1.34 2000/12/22 13:36:12 ulammers Exp $

#include <iostream>
#include <cmath>
#include <cstdio>

#include "param.h"
#include "Dal.h"
#include "errstr.h"
#include <ExprColumnGenerator.h>
#include "Extractor.h"
#include "DataSubspace.h"
#include "DataFilter.h"
#include "xmmexposure.h"

void Spec::make(bool writeDss){

  // Read in the needed parameters

  int IbinSize = intParameter("spectralbinsize");

  // Get pointers to the needed columns

  Column *eDatCol;
  Column *eCol;

  string ecolumn = stringParameter("energycolumn");
  if( _intable->hasColumn(ecolumn) ){
    eCol = _intable->column(ecolumn);

    if( eCol->dataType()!=Dal::Int8 && eCol->dataType()!=Dal::Int16 && 
	eCol->dataType()!=Dal::Int32 )
      errstr << "Energy column is not integer-valued.  All values will be "
             << "truncated to integers before accumulation of spectrum."
	     << Sas::warning("SpecIntColumn");

    eDatCol = reType(eCol, Dal::Int32);
  }else{
    strstream newName;
    newName << "E" << _intable << ends;
    ExprColumnGenerator colGen(ecolumn, newName.str(), Dal::Int32);
    colGen.doIt(*_intable, *_tempTable);
    eCol = eDatCol = _intable->column(ecolumn);
  }

  // Determine the energy range

  vector<real64> limits(2);
  GetLimits(eCol, "specchannelmin", "specchannelmax", booleanParameter("withspecranges"), limits);

  vector<int32> useRange(2);
  useRange[0] = (int32)limits[0];
  useRange[1] = (int32)limits[1];
  vector<int32> tlRange(2);

  if( useRange[1] == useRange[0] )
    errstr << "The energy column for extracting the spectrum has "
	   << "the same value throughout, or the upper and lower limits "
	   << "for this column are equal. An empty spectrum will "
	   << "be created." << Sas::warning("BadSpecLimits"); 

  if( useRange[1] < useRange[0] )
    errstr << "Upper spectral limit < lower limit.  Cannot continue." 
	   << Sas::error("BadSpecLimits");

  unsigned long nRows = (unsigned long) ceil( ((float)(useRange[1]-useRange[0]+1))/IbinSize );

  // Now initialize the output spectrum

  Table * outTable = _dataSet->addTable("SPECTRUM", nRows);

  _copyGlobalAttribs(outTable);

  // Write some keywords

  *outTable->addAttribute( "HDUCLASS",
			   "Format conforms to OGIP/GSFC conventions")
    = "OGIP";
  *outTable->addAttribute( "HDUCLAS1",
			    "File contains a spectrum")
    = "SPECTRUM";
  
  *outTable->addAttribute( "HDUCLAS2",
			   "File contains gross counts")
    = "TOTAL";

  *outTable->addAttribute( "HDUCLAS3",
			   "Spectrum is stored as counts")
    = "COUNT";

  *outTable->addAttribute( "HDUVERS1",
			   "Version of format")
    = "1.1.0";

  *outTable->addAttribute( "AREASCAL",
			   "Nominal scaling factor for data")
    = 1.0;

  *outTable->addAttribute( "BACKSCAL",
			   "Scaling factor for background")
    = 1.0;

  *outTable->addAttribute( "CORRSCAL",
			   "Nominal scaling factor for correction file")
    = 1.0;

  *outTable->addAttribute( "POISSERR",
			   "Poisson errors appropriate")
    = true;

  *outTable->addAttribute( "SYS_ERR",
			   "Global systematic error")
    = 0;

  *outTable->addAttribute( "CHANTYPE","Type of channel data") = eCol->name();

  *outTable->addAttribute( "DETCHANS",
			   "Total number of detector channels available")
    = (int)nRows;

  *outTable->addAttribute( "QUALITY",
			   "All channels have good quality")
    = 0;

  *outTable->addAttribute( "GROUPING",
			   "No data grouping done")
    = 0;

  *outTable->addAttribute( "SPECDELT",
			   "Spectral channel size, ie the binning factor")
    = IbinSize;

  *outTable->addAttribute( "SPECPIX",
			   "The rebinned channel correspondsing to SPECVAL")
    = 0;

  *outTable->addAttribute( "SPECVAL",
			   "Original chan value at center of rebinned chan")
    = useRange[0] - 0.5 + float(IbinSize)/2;

  // Make the necessary columns.
  
  Column * chanCol = outTable->addColumn("CHANNEL", Dal::Int16, "");
  ColumnData * cChanDat = chanCol->data();
  int16 * chanDat = cChanDat->int16Data();

  *chanCol->addAttribute("TLMIN") = 0;

  *chanCol->addAttribute("TLMAX") = (int)nRows-1;

  ColumnData * cCntDat = outTable->addColumn("COUNTS", Dal::Int32, "", "count")->data();
  int32 * cntDat = cCntDat->int32Data();
  
  for(unsigned i=0; i<nRows; i++){
    chanDat[i] = i;
    cntDat[i] = 0;
  }

  delete cChanDat;

  // Set up type converted columns

  Column *flagDatCol;
  ColumnData *cFlagData = 0, *cEData = eDatCol->data();
  int32 index, *flagData = 0, *eData = cEData->int32Data();

  if(! _destruct){
    flagDatCol = reType(_intable->column(_flagCol), Dal::Int32);
    cFlagData = flagDatCol->data();
    flagData = cFlagData->int32Data();
  }

  // Accumlate the spectrum

  unsigned long mRows = _intable->rows();
  for(unsigned long irow=0; irow<mRows; irow++){
    if( ! _destruct && (( flagData[irow] & _flagBit) == 0) ) continue;

    index = (eData[irow]-useRange[0])/IbinSize;
    if( index >= 0 && index < (int32)nRows ) cntDat[index]++;
  }
 
  delete cEData;

  delete cCntDat;

  if(! _destruct) delete cFlagData;

  // Write data subspace info
  if (writeDss) {
     DataSubspace dss(_intable);
     outTable << dss;
  }

  // Calculate exposure keyword

  _getExposure(outTable);

  // Done
}
