//	xmmexposure.cc --- implementation of the XmmExposure class
//
//	DESCRIPTION:
//
//	AUTHOR:
//		D. Hinshaw
//

#include <cstdio>
#include <unistd.h>

#include "CalServer.h"
#include "Xmm.h"

#include "param.h"
#include "errstr.h"
#include "Dal.h"
#include "DataSubspace.h"
#include "DataFilter.h"
#include "xmmexposure.h"
#include "OGIPgti.h"

XmmExposure::XmmExposure(Table *eventTable) : _eventTable(eventTable) {

  _set = _eventTable->parent();
  _instrument = Xmm::INSTRUMENT_NONE;

  // Do some checks to make sure we can do exposure correction on this data set

  if( !_set->hasAttribute("TELESCOP") ||
      _set->attribute("TELESCOP")->asString() != "XMM" ){
    errstr << "Can only update exposure information on XMM data." << Sas::warning("NotXMM");
    return;
  }

  _instrument = Xmm::toInstrument(_set->attribute("INSTRUME")->asString());

  // Get the list of CCDs with exposure extensions

  for(int i=1; i<=17; i++){
    strstream extName;
    extName << "EXPOSU";
    extName.width(2);
    extName.fill('0');
    extName << i << ends;

    if( _set->hasBlock(extName.str()) ) _ccdNums.push_back(i);
  }

  calServer.setInstrument(_instrument);

  if( _instrument==Xmm::EMOS1 || _instrument==Xmm::EMOS2 ||
      _instrument==Xmm::EPN ){
    MiscDataServer *miscData = calServer.getAtom(miscData);
    _targetCCD = (int)miscData->get("OPTICS_CCD");
    _getGtiLists();
  }else if( _instrument==Xmm::RGS1 || _instrument==Xmm::RGS2 ){

    // Check for any time based filters

    DataSubspace dss(_eventTable);
    unsigned ncomps = dss.components();

    for(unsigned icomp=0; icomp<ncomps; icomp++){
      DScomp * comp = dss.component(icomp);
      if( comp->hasFilter("TIME", DSS::RangeT) || comp->hasFilter("TIME", DSS::RangeBlockT) ||
	  comp->hasFilter("TIME", DSS::GtiT) ){
	errstr << "You have used the TIME column to filter an RGS events list.  "
	       << "This will invalidate any exposure information in the produced products."
	       << Sas::warning("RgsTimeFilter");
	return;
      }
    }
  }
  
  return;
}


void XmmExposure::updateOntime() {

  if(_instrument==Xmm::INSTRUMENT_NONE ||
	 _instrument==Xmm::RGS1 ||
	 _instrument==Xmm::RGS2) return;

  // Sum up all the GTIs to get the ontime    

  for(unsigned iccd=0; iccd<_ccdNums.size(); iccd++){
    double ontime = 0;
    assert(_gtiLists.size() > iccd);
    for(unsigned igti=0; igti<_gtiLists[iccd].size(); igti++)
      ontime += _gtiLists[iccd][igti].stop() - _gtiLists[iccd][igti].start();

    // Write the new ontime keywords.

    strstream keyName;
    keyName << "ONTIME";
    keyName.width(2);
    keyName.fill('0');
    keyName << _ccdNums[iccd]  << ends;
    
    *_eventTable->addAttribute(keyName.str(), "Sum of GTIs for a particular CCD") 
      = ontime;
  }

  // Write ONTIME keyword for target CCD

  strstream keyName;
  keyName << "ONTIME";
  keyName.width(2);
  keyName.fill('0');
  keyName << _targetCCD  << ends;

  if( _eventTable->hasAttribute(keyName.str()) )
    *_eventTable->addAttribute("ONTIME", "Sum of GTIs for central CCD") 
      = _eventTable->attribute(keyName.str())->asReal();
  else if( _ccdNums.size()>0 ){
    errstr << "Ontime for target "
	   << "CCD (number " << _targetCCD << ") is not available. "
	   << "Setting ONTIME to zero."
	   << Sas::warning("NoTargetOntime");
    *_eventTable->addAttribute("ONTIME", "Sum of GTIs for central CCD") = 0;
  }

  return;
}


void XmmExposure::updateLivetime(){

  if(_instrument==Xmm::INSTRUMENT_NONE ||
	 _instrument==Xmm::RGS1 ||
	 _instrument==Xmm::RGS2) return;

  for(unsigned i=0; i<_ccdNums.size(); i++){
    strstream keyName;
    keyName << "LIVETI";
    keyName.width(2);
    keyName.fill('0');
    keyName << _ccdNums[i] << ends;
    
    *_eventTable->addAttribute(keyName.str(), "Live time for a particular CCD") = _getLiveTime(i);
  }

  // Write LIVETIME keyword for the central CCD

  strstream keyName;
  keyName << "LIVETI";
  keyName.width(2);
  keyName.fill('0');
  keyName << _targetCCD << ends;
  
  if( _eventTable->hasAttribute(keyName.str()) )
    *_eventTable->addAttribute("LIVETIME", "Live time for the central CCD") 
      = _eventTable->attribute(keyName.str())->asReal();
  else if( _ccdNums.size()>0 ){
    errstr << "Live time for target "
	   << "CCD (number " << _targetCCD << ") is not available. "
	   << "Setting LIVETIME to zero."
	   << Sas::warning("NoTargetLivetime");
    *_eventTable->addAttribute("LIVETIME", "Live time for the central CCD") = 0;
  }

  return;
}


void XmmExposure::_getGtiLists() {

  // Initialize Gti lists with tstart and tstop of EXPOSUnn extensions

  for(unsigned i=0; i<_ccdNums.size(); i++){
    strstream extName;
    extName << "EXPOSU";
    extName.width(2);
    extName.fill('0');
    extName << _ccdNums[i] << ends;
 
    Block* expBlock = _set->block(extName.str());
    Gti expGti(expBlock->attribute("TSTART")->asReal(),
	       expBlock->attribute("TSTOP")->asReal());
 
    GtiList tempList;
    tempList.push_back(expGti);
    _gtiLists.push_back(tempList);
  }

  // Step thru each component of the DSS. Each one should have a CCDNR filter.  Otherwise there
  // should be no CCDNR filters.

  DataSubspace dss(_eventTable);
  unsigned ncomps = dss.components();

  if(ncomps>0){

    bool noCcds = true;
    for(unsigned icomp=0; icomp<ncomps; icomp++){
      DScomp * comp = dss.component(icomp);
      if( comp->hasFilter("CCDNR", DSS::RangeT) || comp->hasFilter("CCDNR", DSS::RangeBlockT) ){
	noCcds = false;
	break;
      }
    }

    if(noCcds){

      // Need to check if TIME filters have been applied in a way that makes sense in terms
      // of exposure correction, ie that different sets of events do not have different
      // exposure corrections.

      DataSubspace dssTemp = dss;
      DScomp * comp0 = dssTemp.component(0);
    
      // Check if we have just global TIME filters, or that all other filters are global.

      bool globalTime = true;
      bool globalOther = true;
      for(unsigned icomp=0; icomp<dss.components(); icomp++){
	for(unsigned ifilter=0; ifilter<dss.component(icomp)->filters(); ifilter++){
	  DSfilter *filter = dss.component(icomp)->filter(ifilter);
	  if( !dssTemp.isGlobal(*filter) ){
	    if(filter->name() == "TIME") globalTime = false;
	    else globalOther = false;
	  }
	}
      }

      if(globalTime){
	while( comp0->hasFilter("TIME", DSS::GtiT) ){
	  DSfilter * filter0 = comp0->filter("TIME", DSS::GtiT);
	  GtiList glist;
	  InOGIPgtiTable oGtiTab(filter0->blockSpec());
	  oGtiTab >> glist;
	  for(unsigned i=0; i<_ccdNums.size(); i++) _gtiLists[i] &= glist;
      
	  for(unsigned ifilter=0; ifilter<comp0->filters(); ifilter++)
	    if( comp0->filter(ifilter) == filter0 ){
	      comp0->deleteFilter(ifilter);
	      break;
	    }
	}

	while( comp0->hasFilter("TIME", DSS::RangeT) ){
	  DSfilter * filter0 = comp0->filter("TIME", DSS::RangeT);
	  GtiList glist;
	  list<Range<double> > ranges = filter0->ranges();
	  list<Range<double> >::const_iterator irange;
	  for(irange=ranges.begin(); irange!=ranges.end(); irange++){
	    double lower, upper;
	    RangeBound lowerB, upperB;
	    irange->values(lower, upper, lowerB, upperB);
	    if( lowerB==Unbounded ) lower=0;
	    if( upperB==Unbounded ) upper=Gti::ETERNITY;
	    Gti gti(lower, upper);
	    glist.push_back(gti);
	  }
	  for(unsigned i=0; i<_ccdNums.size(); i++) _gtiLists[i] &= glist;
	  
	  for(unsigned ifilter=0; ifilter<comp0->filters(); ifilter++)
	    if( comp0->filter(ifilter) == filter0 ){
	      comp0->deleteFilter(ifilter);
	      break;
	    }
	}
	
      }else{
	// TIME filters are not all global, so check if all other filters are global.
	// If not, warn the user that he is on his own.
	if( !globalOther )
	  errstr << "A filtering expression has been specified which may not allow for "
		 << "calculation of a uniform exposure for all events.  Please ensure that "
		 << "all TIME related filters are applied uniformly."
		 << Sas::warning("ExpBadDssGlobal");
      
	GtiList glist;
	for(unsigned icomp=0; icomp<dssTemp.components(); icomp++){
	  GtiList compGlist(true);
	  DScomp * comp = dssTemp.component(icomp);
	  for(unsigned ifilter=0; ifilter<comp->filters(); ifilter++){
	    DSfilter *filter = comp->filter(ifilter);
	    if(filter->name()=="TIME"){
	      GtiList filterGlist;
	      if(filter->type()==DSS::GtiT){
		InOGIPgtiTable oGtiTab(filter->blockSpec());
		oGtiTab >> filterGlist;
		comp->deleteFilter(ifilter);
	      }else if(filter->type()==DSS::RangeT){
		list<Range<double> > ranges = filter->ranges();
		list<Range<double> >::const_iterator irange;
		for(irange=ranges.begin(); irange!=ranges.end(); irange++){
		  double lower, upper;
		  RangeBound lowerB, upperB;
		  irange->values(lower, upper, lowerB, upperB);
		  if( lowerB==Unbounded ) lower=0;
		  if( upperB==Unbounded ) upper=Gti::ETERNITY;
		  Gti gti(lower, upper);
		  filterGlist.push_back(gti);
		}
		comp->deleteFilter(ifilter);
	      }
	      compGlist &= filterGlist;
	    }
	  }
	  glist |= compGlist;
	}
	for(unsigned i=0; i<_ccdNums.size(); i++) _gtiLists[i] &= glist;
      }

      for(unsigned icomp=0; icomp<dssTemp.components(); icomp++){
	DScomp *comp = dssTemp.component(icomp);
	for(unsigned ifilter=0; ifilter<comp->filters(); ifilter++){
	  if( comp->filter(ifilter)->name() == "TIME" ){
	    errstr << "Time filter was encountered which cannot (and will not) be "
		   << "included in exposure calculations: " << comp->filter(ifilter)->selectExpression()
		   << Sas::warning("ExpTimeFilter");
	    return;
	  }
	}
      }
    
    }else{

      // Have per-CCD components
      
      // Create a table of CCD numbers for test filtering

      DataSet *tempSet = dataSetServer->open("EvTempCcd.fits", DataSet::Temp);
      Table * ccdTable = tempSet->addTable("ccdTable", _ccdNums.size());
      Column * ccdCol = ccdTable->add(_eventTable->column("CCDNR"));
      
      ColumnData * ccdColDat = ccdCol->data();
      int8 * ccdDat = ccdColDat->int8Data();
      for(unsigned i=0; i<_ccdNums.size(); i++) ccdDat[i] = _ccdNums[i];
      
      for(unsigned icomp=0; icomp<ncomps; icomp++){
	DScomp *comp = dss.component(icomp);
	
	DSS::DSaxes axes(1, "CCDNR");
	DataFilter filter(comp->selectExpression(axes));
	vector<DataFilter::RowRange> rowranges;
	if( filter.doSelect(*ccdTable, rowranges) != 1 ){
	  delete ccdColDat;
	  dataSetServer->close(tempSet);
	  errstr << "Event list does not contain filtering information with the expected structure. "
		 << "Calculated exposure information is not reliable." << Sas::warning("ExpBadDSSccd");
	  return;
	}
	
	unsigned long iccd = rowranges[0].startRow;
	ccdDat[iccd] = 99;
	
	// Now step thru the filters, looking for TIME filters
	
	if( !comp->hasFilter("TIME", DSS::GtiT) ) continue;
	
	for(unsigned ifilter=0; ifilter<comp->filters(); ifilter++){
	  DSfilter *filter = comp->filter(ifilter);
	  if( filter->name()=="TIME" ){
	    GtiList glist;
	    if( filter->type()==DSS::GtiT ){
	      InOGIPgtiTable oGtiTab(filter->blockSpec());
	      oGtiTab >> glist;
	      _gtiLists[iccd] &= glist;
	    }else if(filter->type()==DSS::RangeT){
	      list<Range<double> > ranges = filter->ranges();
	      list<Range<double> >::const_iterator irange;
	      for(irange=ranges.begin(); irange!=ranges.end(); irange++){
		double lower, upper;
		RangeBound lowerB, upperB;
		irange->values(lower, upper, lowerB, upperB);
		if( lowerB==Unbounded ) lower=0;
		if( upperB==Unbounded ) upper=Gti::ETERNITY;
		Gti gti(lower, upper);
		glist.push_back(gti);
	      }
	      _gtiLists[iccd] &= glist;
	    }else{
	      errstr << "Time filter encountered which cannot (and will not) be "
		     << "included in exposure calculations: " << filter->selectExpression()
		     << Sas::warning("ExpTimeFilter");
	    }
	  }
	}
      }
      
      delete ccdColDat;
      dataSetServer->close(tempSet);
    }
  }

  for(unsigned i=0; i<_ccdNums.size(); i++){
      _gtiLists[i].reduce();
      _gtiLists[i].sortIt();
  }

  // If requested, filter the EXPOSURE extensions with their corresponding GTI.
  
  if( booleanParameter("keepfilteroutput") && booleanParameter("filterexposure") ){
    string tempSetName = tempnam(0,0);
    for(unsigned i=0; i<_ccdNums.size(); i++){
      strstream extName;
      extName << "EXPOSU";
      extName.width(2);
      extName.fill('0');
      extName << _ccdNums[i] << ends;

      if( _gtiLists[i].size() == 0 ){
	Table * expTab = _set->table(extName.str());
	expTab->deleteRows(0, expTab->rows());
      }else{
	DataSet *tempSet = dataSetServer->open(tempSetName, DataSet::Create);
	OutOGIPgtiTable tempgti(tempSet);
	tempgti << _gtiLists[i];
	dataSetServer->close(tempSet);

	DataFilter filter( "gti(" + tempSetName + ",TIME)" );
	filter.doFilter(*_set->table(extName.str()));

	unlink(tempSetName.c_str());
      }
    }
  }

  return;
}


real64 XmmExposure::_getLiveTime(unsigned ccdRef){

    unsigned ccdNr = _ccdNums[ccdRef];

    strstream extName;
    extName << "EXPOSU";
    extName.width(2);
    extName.fill('0');
    extName << ccdNr << ends;

    Table * expTab = _set->table(extName.str());

    real32 delTime = 0.;
    bool hasTimeDelCol = true;
    if( !expTab->hasColumn("TIMEDEL") ){
      hasTimeDelCol = false;
      delTime = expTab->attribute("TIMEDEL")->asReal();
    }

    ColumnData * timeCdat = expTab->column("TIME")->data();
    ColumnData * fracExpCdat = expTab->column("FRACEXP")->data();
    ColumnData * delTimeCdat = 0;
    if(hasTimeDelCol) delTimeCdat = expTab->column("TIMEDEL")->data();
    
    real64 * timeDat = timeCdat->real64Data();
    real32 * fracExpDat = fracExpCdat->real32Data();
    real32 * delTimeDat = 0;
    if(hasTimeDelCol) delTimeDat = delTimeCdat->real32Data();

    real64 timePixR = 0.5;
    if( expTab->hasAttribute("TIMEPIXR") ) 
      timePixR = expTab->attribute("TIMEPIXR")->asReal();

    GtiList gtilist = _gtiLists[ccdRef];
    const Gti * gti;

    real64 livetime = 0;
    unsigned long mRows = expTab->rows();
    for(unsigned long irow=0; irow<mRows; irow++){
      if(hasTimeDelCol) delTime = delTimeDat[irow];
      if( (gti=gtilist.inside(timeDat[irow]-delTime*timePixR)) != NULL && 
	  gti->stop() > timeDat[irow] + delTime*timePixR )
	livetime += delTime*fracExpDat[irow];
    }

    delete timeCdat;
    if(hasTimeDelCol) delete delTimeCdat;
    delete fracExpCdat;

    return livetime;
}


real64 XmmExposure::liveTime(unsigned ccdNr){
  unsigned iccdRef;
  for(iccdRef=0; iccdRef<_ccdNums.size(); iccdRef++)
    if( ccdNr == _ccdNums[iccdRef] ) break;

  if( iccdRef == _ccdNums.size() ) return 0;

  return _getLiveTime(iccdRef);
}
